import uuid
import webbrowser
from abc import ABC, abstractmethod
from http.server import HTTPServer
from typing import Optional, Tuple

import click
import pkce
import requests
from sym.shared.cli.errors import CliError

from sym.flow.cli.errors import LoginError
from sym.flow.cli.helpers.constants import DEFAULT_SCOPES
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.login.handler import make_handler
from sym.flow.cli.models import AuthToken, Organization, parse_auth_token


class LoginFlow(ABC):
    @abstractmethod
    def login(self, options: GlobalOptions, org: Organization) -> AuthToken:
        pass


class BrowserRedirectFlow(LoginFlow):
    """
    Issue an authorization code request by opening the browser to a particular
    URL. Start a server and wait for the user to successfully login. The browser will
    redirect a successful login back to the locally running server, with a request
    body containing an auth code. With the auth code, obtain an access and refresh token.

    Implements the auth code flow described here:
    https://www.altostra.com/blog/cli-authentication-with-auth0
    """

    port: int

    def __init__(self, port: int):
        self.port = port
        self.browser_login_params = None

    def login(self, options: GlobalOptions, org: Organization) -> Optional[AuthToken]:
        auth_code, code_verifier = self.get_auth_code(options, org)
        if not auth_code:
            return
        return self.get_token_from_code(options, org, auth_code, code_verifier)

    @property
    def redirect_url(self) -> str:
        return f"http://localhost:{self.port}/callback"

    def _new_server(self, options: GlobalOptions, handler):
        for port in range(self.port, self.port + 10):  # try a few
            try:
                return HTTPServer(("localhost", port), handler)
            except OSError as e:
                if e.errno not in (48, 98):  # Address already in use
                    raise
                options.dprint(f"Port {port} in use, trying to increment...")

    def wait_for_code(self, state: str, options: GlobalOptions) -> Optional[str]:
        """
        Starts a server and waits for a request that contains an auth code.
        """
        handler = make_handler(options, state)
        with self._new_server(options, handler) as httpd:
            options.dprint(f"Waiting for an auth code at {self.redirect_url}")
            httpd.handle_request()

            # Bubble up any error message received during handling of the request from the HTTPServer.
            if error_message := getattr(handler, "error_message", None):
                # If we know the problem is that the user just signed up (and therefore their email is not validated),
                # give them a non-error message.
                if "complete sign up" in error_message:
                    click.secho(error_message, fg="cyan")
                    return
                else:
                    # Otherwise, raise the error message that we got from the HTTP response.
                    raise CliError(error_message)

            # We didn't get an auth_code, but the HTTPServer didn't register any errors.
            if not (auth_code := getattr(handler, "auth_code", None)):
                raise CliError("Unable to get access token from Auth0. Please try again.")

            return auth_code

    def gen_state(self) -> str:
        return str(uuid.uuid4())

    def gen_browser_login_params(
        self, options: GlobalOptions, org: Organization, force: bool = False
    ) -> str:
        if force or self.browser_login_params is None:
            scopes_str = " ".join(DEFAULT_SCOPES)
            state = self.gen_state()
            code_verifier = pkce.generate_code_verifier(length=128)
            code_challenge = pkce.get_code_challenge(code_verifier)
            query_params = "&".join(
                [
                    "response_type=code",
                    f"client_id={org.client_id}",
                    "code_challenge_method=S256",
                    f"code_challenge={code_challenge}",
                    f"redirect_uri={self.redirect_url}",
                    f"scope={scopes_str}",
                    f"audience=https://api.symops.com",
                    f"state={state}",
                    "prompt=login",
                ]
            )
            url = f"{options.auth_url}/authorize?{query_params}"
            self.browser_login_params = (url, state, code_verifier)

        return self.browser_login_params

    def get_auth_code(self, options: GlobalOptions, org: Organization) -> Tuple[str, str]:
        """
        Perform the Auth0 code flow by opening the browser with a URL containing query parameters
        that orchestrate the flow. When the user logs in successfully, the browser will follow the
        redirect and issue a request back to the CLI with an auth code.

        Returns an auth code and a PKCE code verifier that both need to be provided to the
        token endpoint when obtaining an auth token.
        """
        (url, state, code_verifier) = self.gen_browser_login_params(options, org)

        webbrowser.open(url)

        auth_code = self.wait_for_code(state, options)
        options.dprint(f"Auth code received: {auth_code}")
        return auth_code, code_verifier

    def get_token_from_code(
        self,
        options: GlobalOptions,
        org: Organization,
        auth_code: str,
        code_verifier: str,
    ) -> AuthToken:
        """
        With the provided auth code and verifier, obtains an auth token.
        """
        data = {
            "grant_type": "authorization_code",
            "client_id": org.client_id,
            "code_verifier": code_verifier,
            "code": auth_code,
            "redirect_uri": self.redirect_url,
        }
        headers = {"content-type": "application/x-www-form-urlencoded"}
        url = f"{options.auth_url}/oauth/token"
        r = requests.post(url, headers=headers, data=data)
        if not r.ok:
            raise LoginError(r.text)

        return parse_auth_token(r.json())



import urllib.parse as parse
from http.server import BaseHTTPRequestHandler
from typing import ClassVar, Type

from ...helpers.global_options import GlobalOptions


class AuthHTTPRequestHandler(BaseHTTPRequestHandler):

    options: ClassVar[GlobalOptions]
    state: ClassVar[str]
    auth_code: ClassVar[str]
    error_message: ClassVar[str]

    def log_message(self, format, *args):
        ...

    def do_GET(self):
        parsed = parse.urlparse(self.path)
        qs = parse.parse_qs(parsed.query)

        if qs.get("error_description"):
            error_message = qs["error_description"][0]
        elif not qs.get("code"):
            error_message = "Missing code in query"
        elif not qs.get("state"):
            error_message = "Missing state in query"
        elif not qs["state"][0] == self.__class__.state:
            self.__class__.options.dprint(
                f"Invalid state. Found: {qs['state'][0]}, expected: {self.__class__.state}."
            )
            error_message = "Invalid state"
        else:
            error_message = None

        if error_message:
            query = parse.urlencode({"message": error_message})
            self.send_response(301)
            self.send_header("Location", f"https://static.symops.com/cli/error?{query}")
            self.end_headers()
            self.__class__.error_message = error_message
            return

        self.send_response(301)
        self.send_header("Location", "https://static.symops.com/cli/login")
        self.end_headers()

        self.__class__.auth_code = qs["code"][0]


def make_handler(options: GlobalOptions, state: str) -> Type[AuthHTTPRequestHandler]:
    class Handler(AuthHTTPRequestHandler):
        ...

    Handler.options = options
    Handler.state = state
    return Handler



import importlib.resources as pkg_resources
import os

import click

from sym.flow.cli.code_generation import (  # import the *package* containing the tf files
    approval,
)
from sym.flow.cli.errors import NonEmptyDirectoryError, NotLoggedInError
from sym.flow.cli.helpers.config import Config
from sym.flow.cli.helpers.global_options import GlobalOptions


@click.command(short_help="Generate starter Terraform code")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.option(
    "--workspace-id",
    prompt="Workspace ID",
)
def init(options: GlobalOptions, workspace_id: str) -> None:
    """In an empty directory, generate the Terraform configuration needed to create a Sym approval-only Flow."""

    if not Config.is_logged_in():
        raise NotLoggedInError()

    if os.listdir("."):
        raise NonEmptyDirectoryError()

    # Note: The impl file is stored as a `.txt` resource because PyOxidizer (the tool used to package symflow CLI)
    # Does NOT support reading `.py` files with `importlib.resources`
    # https://github.com/indygreg/PyOxidizer/issues/237
    #
    # However, we don't care about reading the source code, we simply need to pull the text file and write it
    # to the filesystem with a `.py` extension. As a workaround, we have stored `impl.py` as `impl.txt` in the
    # code_generation.approval package so that we can read it with importlib.resources.
    static_files = [("impl.txt", "impl.py"), ("versions.tf", "versions.tf")]

    # These files don't need any values substituted in them.
    for resource_name, output_file in static_files:
        template = pkg_resources.read_text(approval, resource_name)
        with open(output_file, "w") as f:
            f.write(template)

    # Fill in the org slug, auto-approve list, and workspace ID variables with values
    org = Config.get_org().slug
    main_tf = pkg_resources.read_text(approval, "main.tf")
    with open("main.tf", "w") as f:
        main_tf = main_tf.replace("${var.sym_org_slug}", org)
        main_tf = main_tf.replace("${var.slack_workspace_id}", workspace_id.upper())

        f.write(main_tf)

    click.secho(
        f"Successfully generated your Sym Terraform configuration! Run the following to apply the Terraform:\n",
        fg="green",
    )
    click.secho(f"terraform init && terraform apply", fg="green")



from unittest.mock import patch

from sym.flow.cli.symflow import symflow as click_command


class TestUsersDelete:
    @patch("sym.flow.cli.helpers.api.SymAPI.delete_user")
    def test_delete(self, mock_delete_user, click_setup):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["users", "delete", "user@symops.io"])
            assert result.exit_code == 0

        mock_delete_user.assert_called_once()

    @patch("sym.flow.cli.helpers.api.SymAPI.delete_user")
    def test_delete_missing_arguments(self, _, click_setup):
        with click_setup() as runner:
            result = runner.invoke(click_command, ["users", "delete"])
            assert result.exit_code == 2



import importlib.resources as pkg_resources
import os

from sym.flow.cli.code_generation import approval
from sym.flow.cli.helpers.config import store_login_config
from sym.flow.cli.symflow import symflow as click_command


class TestInit:
    def test_init(self, click_setup, test_org, auth_token):
        with click_setup() as runner:
            store_login_config("implementer@symops.io", test_org, auth_token)

            # Directory is empty to begin
            assert not os.listdir(".")

            result = runner.invoke(click_command, ["init", "--workspace-id", "T123ABC"])
            assert result.exit_code == 0
            assert (
                "Successfully generated your Sym Terraform configuration! Run the following to apply the Terraform:"
                in result.output
            )
            assert "terraform init && terraform apply" in result.output

            # Check that we wrote all of the Terraform files to the directory
            files = os.listdir(".")
            assert "main.tf" in files
            assert "impl.py" in files
            assert "versions.tf" in files

            # For the static files, make sure they are written as-is
            expected_impl = pkg_resources.read_text(approval, "impl.txt")
            with open("impl.py", "r") as f:
                written_impl = f.read()
            assert written_impl == expected_impl

            expected_file = pkg_resources.read_text(approval, "versions.tf")
            with open("versions.tf", "r") as f:
                written_file = f.read()
            assert written_file == expected_file

            # For main.tf, make sure that all of the expected things are filled in.
            expected_main_tf = pkg_resources.read_text(approval, "main.tf")
            expected_main_tf = expected_main_tf.replace(
                "${var.sym_org_slug}", test_org.slug
            )
            expected_main_tf = expected_main_tf.replace(
                "${var.slack_workspace_id}", "T123ABC"
            )

            with open("main.tf", "r") as f:
                written_main_tf = f.read()

            assert written_main_tf == expected_main_tf

    def test_init_prompts(self, click_setup, test_org, auth_token):
        with click_setup() as runner:
            store_login_config("implementer@symops.io", test_org, auth_token)
            result = runner.invoke(click_command, ["init"], input="T123ABC\n")
            assert result.exit_code == 0

            assert (
                "Successfully generated your Sym Terraform configuration! Run the following to apply the Terraform:"
                in result.output
            )
            assert "terraform init && terraform apply" in result.output

    def test_init_errors_non_empty(self, click_setup, test_org, auth_token):
        with click_setup() as runner:
            store_login_config("implementer@symops.io", test_org, auth_token)

            # Write a random file so the directory isn't empty
            with open("foo.txt", "w") as f:
                f.write("Hello")

            # Directory is not empty
            assert os.listdir(".")

            result = runner.invoke(click_command, ["init", "--workspace-id", "T123ABC"])
            assert result.exit_code != 0
            assert (
                "The current directory is not empty! `symflow init` can only be run in an empty directory."
                in result.output
            )



from contextlib import ExitStack
from unittest.mock import patch

import pytest

from sym.flow.cli.symflow import symflow as click_command
from sym.flow.cli.tests.factories.users import UserFactory


@patch("sym.flow.cli.helpers.rest.SymRESTClient.post")
class TestUsersSetRole:
    @pytest.fixture
    def user(self):
        yield UserFactory.create()

    @pytest.fixture(autouse=True)
    def mocks(self, user):
        with ExitStack() as stack:
            stack.enter_context(
                patch("sym.flow.cli.helpers.api.SymAPI.get_user", return_value=user)
            )
            yield

    @pytest.mark.parametrize("role", ["admin", "member", "guest"])
    def test_set_role_with_option(self, mock_post, click_setup, user, role):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["users", "set-role", user.sym_email, "--role", role],
            )
            assert result.exit_code == 0

        mock_post.assert_called_once_with(
            f"users/{user.id}/set-role", data={"role": role}
        )

    @pytest.mark.parametrize("role", ["admin", "member", "guest"])
    @patch("inquirer.list_input", side_effect=["admin", "member", "guest"])
    def test_set_role_prompt(self, mock_inquirer, mock_post, click_setup, user, role):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["users", "set-role", user.sym_email],
            )
            assert result.exit_code == 0

        mock_inquirer.assert_called_once_with(
            "Which role?", choices=["admin", "guest", "member"]
        )

    def test_set_role_invalid(self, user, click_setup):
        with click_setup() as runner:
            result = runner.invoke(
                click_command,
                ["users", "set-role", user.sym_email, "--role", "boo"],
            )
            assert result.exit_code != 0
            assert "Invalid input: 'boo'" in result.output
            assert "Try one of: admin, guest, member" in result.output
